
# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false
# pyright: reportUnknownArgumentType=false
# pyright: reportMissingTypeStubs=false
# pyright: reportAssignmentType=false

# Imports
import tensorflow as tf
from keras.losses import Loss


class NextGenerationLoss(Loss):
	""" Next Generation Loss with alpha = 2.4092.

	Sources:
	- Code: https://github.com/ZKI-PH-ImageAnalysis/Next-Generation-Loss/blob/main/NGL_torch.py
	- Next Generation Loss Function for Image Classification: https://arxiv.org/pdf/2404.12948
	"""

	def __init__(self, alpha: float = 2.4092, name: str = "ngl_loss"):
		""" Initialize the Next Generation Loss.

		Args:
			alpha    (float):  The alpha parameter.
			name     (str):    The name of the loss function.
		"""
		super().__init__(name=name)
		self.name: str = name
		""" The name of the loss function. """
		self.alpha: float = alpha
		""" The alpha parameter. """

	def call(self, y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
		""" Compute the NGL loss.

		Args:
			y_true (tf.Tensor): The true labels.
			y_pred (tf.Tensor): The predicted labels.
		Returns:
			tf.Tensor: The computed NGL loss.
		"""
		# Cast to float32
		y_pred = tf.cast(y_pred, tf.float32)
		y_true = tf.cast(y_true, tf.float32)

		# Apply softmax to predictions
		y_pred = tf.nn.softmax(y_pred, axis=-1)

		# Compute the NGL loss using the alpha parameter (default 2.4092)
		loss: tf.Tensor = tf.reduce_mean(
			tf.math.exp(self.alpha - y_pred - y_pred * y_true) -
			tf.math.cos(tf.math.cos(tf.math.sin(y_pred)))
		)
		return loss




""" ResNet models implementation.

This module provides wrapper classes for the ResNet family of models from the Keras applications.
It includes both ResNetV2 models with pre-activation residual blocks and ResNetRS
(ResNet with Revisited Scaling) models that offer improved performance
through various scaling techniques.

Available models:

- ResNetV2 family: Improved ResNet architectures with pre-activation blocks
    - ResNet50V2
    - ResNet101V2
    - ResNet152V2
- ResNetRS family: ResNet models with revisited scaling for better efficiency
    - ResNetRS50
    - ResNetRS101
    - ResNetRS152
    - ResNetRS200
    - ResNetRS270
    - ResNetRS350
    - ResNetRS420

All models support transfer learning from ImageNet pre-trained weights.
"""
# pyright: reportUnknownVariableType=false
# pyright: reportMissingTypeStubs=false

# Imports
from __future__ import annotations

from keras.models import Model
from keras.src.applications.resnet_rs import ResNetRS50 as ResNetRS50_keras
from keras.src.applications.resnet_rs import ResNetRS101 as ResNetRS101_keras
from keras.src.applications.resnet_rs import ResNetRS152 as ResNetRS152_keras
from keras.src.applications.resnet_rs import ResNetRS200 as ResNetRS200_keras
from keras.src.applications.resnet_rs import ResNetRS270 as ResNetRS270_keras
from keras.src.applications.resnet_rs import ResNetRS350 as ResNetRS350_keras
from keras.src.applications.resnet_rs import ResNetRS420 as ResNetRS420_keras
from keras.src.applications.resnet_v2 import ResNet50V2 as ResNet50V2_keras
from keras.src.applications.resnet_v2 import ResNet101V2 as ResNet101V2_keras
from keras.src.applications.resnet_v2 import ResNet152V2 as ResNet152V2_keras

from ....decorators import simple_cache
from ..base_keras import BaseKeras
from ..model_interface import CLASS_ROUTINE_DOCSTRING, MODEL_DOCSTRING


# Classes
class ResNet50V2(BaseKeras):
	def _get_base_model(self) -> Model:
		return ResNet50V2_keras(include_top=False, classes=self.num_classes)

class ResNet101V2(BaseKeras):
	def _get_base_model(self) -> Model:
		return ResNet101V2_keras(include_top=False, classes=self.num_classes)

class ResNet152V2(BaseKeras):
	def _get_base_model(self) -> Model:
		return ResNet152V2_keras(include_top=False, classes=self.num_classes)

class ResNetRS50(BaseKeras):
	def _get_base_model(self) -> Model:
		return ResNetRS50_keras(include_top=False, classes=self.num_classes)

class ResNetRS101(BaseKeras):
	def _get_base_model(self) -> Model:
		return ResNetRS101_keras(include_top=False, classes=self.num_classes)

class ResNetRS152(BaseKeras):
	def _get_base_model(self) -> Model:
		return ResNetRS152_keras(include_top=False, classes=self.num_classes)

class ResNetRS200(BaseKeras):
	def _get_base_model(self) -> Model:
		return ResNetRS200_keras(include_top=False, classes=self.num_classes)

class ResNetRS270(BaseKeras):
	def _get_base_model(self) -> Model:
		return ResNetRS270_keras(include_top=False, classes=self.num_classes)

class ResNetRS350(BaseKeras):
	def _get_base_model(self) -> Model:
		return ResNetRS350_keras(include_top=False, classes=self.num_classes)

class ResNetRS420(BaseKeras):
	def _get_base_model(self) -> Model:
		return ResNetRS420_keras(include_top=False, classes=self.num_classes)


# Docstrings
for model in [
	ResNet50V2, ResNet101V2, ResNet152V2,
	ResNetRS50, ResNetRS101, ResNetRS152, ResNetRS200,
	ResNetRS270, ResNetRS350, ResNetRS420
]:
	model.__doc__ = MODEL_DOCSTRING.format(model=model.__name__)
	model.class_routine = simple_cache(model.class_routine)
	model.class_routine.__doc__ = CLASS_ROUTINE_DOCSTRING.format(model=model.__name__)




""" SqueezeNet model implementation.

This module provides a wrapper class for the SqueezeNet model, a lightweight CNN architecture
that achieves AlexNet-level accuracy with 50x fewer parameters and a model size of less than 0.5MB.
SqueezeNet uses "fire modules" consisting of a squeeze layer with 1x1 filters followed by an
expand layer with a mix of 1x1 and 3x3 convolution filters.

Available models:
- SqueezeNet: Compact model with excellent performance-to-parameter ratio

The model supports transfer learning from ImageNet pre-trained weights.
"""
# pyright: reportUnknownArgumentType=false
# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false
# pyright: reportMissingTypeStubs=false

# Imports
from __future__ import annotations

from typing import Any

from keras import backend
from keras.layers import (
	Activation,
	Convolution2D,
	Dropout,
	GlobalAveragePooling2D,
	GlobalMaxPooling2D,
	Input,
	MaxPooling2D,
	concatenate,
)
from keras.models import Model
from keras.utils import get_file, get_source_inputs

from ....decorators import simple_cache
from ..base_keras import BaseKeras
from ..model_interface import CLASS_ROUTINE_DOCSTRING, MODEL_DOCSTRING

# Constants
SQ1X1: str = "squeeze1x1"

WEIGHTS_PATH = "https://github.com/rcmalli/keras-squeezenet/releases/download/v1.0/squeezenet_weights_tf_dim_ordering_tf_kernels.h5"
WEIGHTS_PATH_NO_TOP = "https://github.com/rcmalli/keras-squeezenet/releases/download/v1.0/squeezenet_weights_tf_dim_ordering_tf_kernels_notop.h5"


# Modular function for Fire Node
def fire_module(x: Any, fire_id: int, squeeze: int = 16, expand: int = 64):
	""" Create a fire module with specified parameters.

	Args:
		x           (Tensor):         Input tensor
		fire_id     (int):            ID for the fire module
		squeeze     (int):            Number of filters for squeeze layer
		expand      (int):            Number of filters for expand layers

	Returns:
		Tensor:                Output tensor from the fire module
	"""
	s_id: str = f"fire{fire_id}"

	if backend.image_data_format() == "channels_first":
		channel_axis: int = 1
	else:
		channel_axis: int = 3

	x = Convolution2D(squeeze, (1, 1), padding="valid", name=f"{s_id}/squeeze1x1")(x)
	x = Activation("relu", name=f"{s_id}/relu_squeeze1x1")(x)

	left = Convolution2D(expand, (1, 1), padding="valid", name=f"{s_id}/expand1x1")(x)
	left = Activation("relu", name=f"{s_id}/relu_expand1x1")(left)

	right = Convolution2D(expand, (3, 3), padding="same", name=f"{s_id}/expand3x3")(x)
	right = Activation("relu", name=f"{s_id}/relu_expand3x3")(right)

	x = concatenate([left, right], axis=channel_axis, name=f"{s_id}/concat")
	return x


# Original SqueezeNet from paper
def SqueezeNet_keras(  # noqa: N802
	include_top: bool = True,
	weights: str = "imagenet",
	input_tensor: Any = None,
	input_shape: tuple[Any, ...] | None = None,
	pooling: str | None = None,
	classes: int = 1000
) -> Model:
	""" Instantiates the SqueezeNet architecture.

	Args:
		include_top            (bool):           Whether to include the fully-connected layer at the top
		weights                (str):            One of `None` or 'imagenet'
		input_tensor           (Tensor):         Optional Keras tensor as input
		input_shape            (tuple):          Optional shape tuple
		pooling                (str):            Optional pooling mode for feature extraction
		classes                (int):            Number of classes to classify images into

	Returns:
		Model:                 A Keras model instance
	"""

	if weights not in {'imagenet', None}:
		raise ValueError(
			"The `weights` argument should be either `None` (random initialization) "
			"or `imagenet` (pre-training on ImageNet)."
		)

	if include_top and weights == 'imagenet' and classes != 1000:
		raise ValueError(
			"If using `weights` as imagenet with `include_top` as true, `classes` should be 1000"
		)

	# Manually handle input shape logic instead of _obtain_input_shape
	default_size: int = 227
	min_size: int = 48
	if backend.image_data_format() == 'channels_first':
		default_shape: tuple[int, int, int] = (3, default_size, default_size)
		if weights == 'imagenet' and include_top and input_shape is not None and input_shape[0] != 3:
			raise ValueError(
				"When specifying `input_shape` and loading 'imagenet' weights, 'channels_first' input_shape "
				"should be (3, H, W)."
			)
	else: # channels_last
		default_shape = (default_size, default_size, 3)
		if weights == 'imagenet' and include_top and input_shape is not None and input_shape[2] != 3:
			raise ValueError(
				"When specifying `input_shape` and loading 'imagenet' weights, 'channels_last' input_shape "
				"should be (H, W, 3)."
			)

	if input_shape is None:
		input_shape = default_shape
	else:
		# Basic validation
		if len(input_shape) != 3:
			raise ValueError("`input_shape` must be a tuple of three integers.")
		if backend.image_data_format() == 'channels_first':
			if input_shape[1] is not None and input_shape[1] < min_size:
				raise ValueError(f"Input size must be at least {min_size}x{min_size}, got `input_shape=`{input_shape}")
			if input_shape[2] is not None and input_shape[2] < min_size:
				raise ValueError(f"Input size must be at least {min_size}x{min_size}, got `input_shape=`{input_shape}")
		else: # channels_last
			if input_shape[0] is not None and input_shape[0] < min_size:
				raise ValueError(f"Input size must be at least {min_size}x{min_size}, got `input_shape=`{input_shape}")
			if input_shape[1] is not None and input_shape[1] < min_size:
				raise ValueError(f"Input size must be at least {min_size}x{min_size}, got `input_shape=`{input_shape}")

	# Handle input tensor
	if input_tensor is None:
		img_input = Input(shape=input_shape)
	else:
		if not backend.is_keras_tensor(input_tensor):
			img_input = Input(tensor=input_tensor, shape=input_shape)
		else:
			img_input = input_tensor

	x = Convolution2D(64, (3, 3), strides=(2, 2), padding='valid', name='conv1')(img_input)
	x = Activation('relu', name='relu_conv1')(x)
	x = MaxPooling2D(pool_size=(3, 3), strides=(2, 2), name='pool1')(x)

	x = fire_module(x, fire_id=2, squeeze=16, expand=64)
	x = fire_module(x, fire_id=3, squeeze=16, expand=64)
	x = MaxPooling2D(pool_size=(3, 3), strides=(2, 2), name='pool3')(x)

	x = fire_module(x, fire_id=4, squeeze=32, expand=128)
	x = fire_module(x, fire_id=5, squeeze=32, expand=128)
	x = MaxPooling2D(pool_size=(3, 3), strides=(2, 2), name='pool5')(x)

	x = fire_module(x, fire_id=6, squeeze=48, expand=192)
	x = fire_module(x, fire_id=7, squeeze=48, expand=192)
	x = fire_module(x, fire_id=8, squeeze=64, expand=256)
	x = fire_module(x, fire_id=9, squeeze=64, expand=256)

	if include_top:
		# It's not obvious where to cut the network...
		# Could do the 8th or 9th layer... some work recommends cutting earlier layers.

		x = Dropout(0.5, name='drop9')(x)

		x = Convolution2D(classes, (1, 1), padding='valid', name='conv10')(x)
		x = Activation('relu', name='relu_conv10')(x)
		x = GlobalAveragePooling2D()(x)
		x = Activation('softmax', name='loss')(x)
	else:
		if pooling == 'avg':
			x = GlobalAveragePooling2D()(x)
		elif pooling == 'max':
			x = GlobalMaxPooling2D()(x)
		elif pooling is None:
			pass
		else:
			raise ValueError("Unknown argument for 'pooling'=" + pooling)

	# Ensure that the model takes into account
	# any potential predecessors of `input_tensor`.
	if input_tensor is not None:
		inputs = get_source_inputs(input_tensor)
	else:
		inputs = img_input

	model = Model(inputs, x, name='squeezenet')

	# load weights
	if weights == 'imagenet':
		if include_top:
			weights_path = get_file('squeezenet_weights_tf_dim_ordering_tf_kernels.h5',
									WEIGHTS_PATH,
									cache_subdir='models')
		else:
			weights_path = get_file('squeezenet_weights_tf_dim_ordering_tf_kernels_notop.h5',
									WEIGHTS_PATH_NO_TOP,
									cache_subdir='models')

		model.load_weights(weights_path)
	return model


# Classes
class SqueezeNet(BaseKeras):
	def _get_base_model(self) -> Model:
		return SqueezeNet_keras(
			include_top=False, classes=self.num_classes, input_shape=(224, 224, 3)
		)


# Docstrings
for model in [SqueezeNet]:
	model.__doc__ = MODEL_DOCSTRING.format(model=model.__name__)
	model.class_routine = simple_cache(model.class_routine)
	model.class_routine.__doc__ = CLASS_ROUTINE_DOCSTRING.format(model=model.__name__)




"""
This module contains the MetricDictionnary class, which provides a dictionary of metric names.

This is often used to log metrics to MLflow and to display them in the console easily.

This class contains the following metrics:

1. Main metrics:

   - Area Under the Curve (AUC)
   - Specificity (True Negative Rate)
   - Recall/Sensitivity (True Positive Rate)
   - Precision (Positive Predictive Value)
   - Negative Predictive Value (NPV)
   - Accuracy
   - F1 Score

2. Confusion matrix metrics:

   - True Negatives (TN)
   - False Positives (FP)
   - False Negatives (FN)
   - True Positives (TP)
   - False Positive Rate
   - False Negative Rate
   - False Discovery Rate
   - False Omission Rate
   - Critical Success Index (Threat Score)

3. F-scores:

   - F-beta Score (where beta is configurable)

4. Matthews correlation coefficient:

   - Matthews Correlation Coefficient (MCC)

5. Optimal thresholds for binary classification:

   - Youden's J statistic
   - Cost-based threshold

6. Average metrics across folds:

   - Mean value of any metric across k-fold cross validation

7. Standard deviation metrics across folds:

   - Standard deviation of any metric across k-fold cross validation
"""

class MetricDictionnary:

	# Main metrics (starting with '1:')
	AUC: str								= "1: Area Under the Curve: AUC"
	SPECIFICITY: str						= "1: Specificity: True Negative Rate"
	RECALL: str								= "1: Recall/Sensitivity: True Positive Rate"
	PRECISION: str							= "1: Precision: Positive Predictive Value"
	NPV: str								= "1: NPV: Negative Predictive Value"
	ACCURACY: str							= "1: Accuracy"
	BALANCED_ACCURACY: str					= "1: Balanced Accuracy"
	F1_SCORE: str							= "1: F1 Score"
	F1_SCORE_NEGATIVE: str					= "1: F1 Score for Negative Class"

	# Confusion matrix metrics (starting with '2:')
	CONFUSION_MATRIX_TN: str				= "2: Confusion Matrix: TN"
	CONFUSION_MATRIX_FP: str				= "2: Confusion Matrix: FP"
	CONFUSION_MATRIX_FN: str				= "2: Confusion Matrix: FN"
	CONFUSION_MATRIX_TP: str				= "2: Confusion Matrix: TP"
	FALSE_POSITIVE_RATE: str				= "2: False Positive Rate"
	FALSE_NEGATIVE_RATE: str				= "2: False Negative Rate"
	FALSE_DISCOVERY_RATE: str				= "2: False Discovery Rate"
	FALSE_OMISSION_RATE: str				= "2: False Omission Rate"
	CRITICAL_SUCCESS_INDEX: str				= "2: Critical Success Index: Threat Score"

	# F-scores (starting with '3:')
	F_SCORE_X: str							= "3: F-X Score"	# X is the beta value

	# Matthews correlation coefficient (starting with '4:')
	MATTHEWS_CORRELATION_COEFFICIENT: str	= "4: Matthews Correlation Coefficient: MCC"

	# Optimal thresholds (starting with '5:')
	OPTIMAL_THRESHOLD_YOUDEN: str			= "5: Optimal Threshold: Youden"
	OPTIMAL_THRESHOLD_COST: str				= "5: Optimal Threshold: Cost"

	# Average metrics across folds (starting with '6:')
	AVERAGE_METRIC: str						= "6: Average METRIC_NAME across folds"

	# Standard deviation metrics across folds (starting with '7:')
	STANDARD_DEVIATION_METRIC: str			= "7: Standard deviation METRIC_NAME across folds"

	# Parameter finder (starting with '8:')
	PARAMETER_FINDER: str					= "8: TITLE: PARAMETER_NAME"





# pyright: reportUnknownVariableType=false
# pyright: reportUnusedImport=false
# pyright: reportArgumentType=false
# pyright: reportCallIssue=false

# Imports
from .common import Any, NDArray, check_image, cv2, np  # noqa: F401


# Functions
def normalize_image(
	image: NDArray[Any],
	a: float | int = 0,
	b: float | int = 255,
	method: int = cv2.NORM_MINMAX,
	ignore_dtype: bool = False,
) -> NDArray[Any]:
	""" Normalize an image to the range 0-255.

	Args:
		image         (NDArray[Any]):  Image to normalize
		a             (float | int): Minimum value (default: 0)
		b             (float | int): Maximum value (default: 255)
		method        (int):         Normalization method (default: cv2.NORM_MINMAX)
		ignore_dtype  (bool):        Ignore the dtype check
	Returns:
		NDArray[Any]: Normalized image

	>>> ## Basic tests
	>>> image = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=np.uint8)
	>>> normalized = normalize_image(image)
	>>> normalized.tolist()
	[[0, 32, 64], [96, 128, 159], [191, 223, 255]]
	>>> normalized.shape == image.shape
	True
	>>> normalized.dtype == image.dtype
	True

	>>> ## Test invalid inputs
	>>> normalize_image("not an image")
	Traceback (most recent call last):
		...
	AssertionError: Image must be a numpy array

	>>> normalize_image(image, a="not an integer")
	Traceback (most recent call last):
		...
	AssertionError: a must be a float or an integer

	>>> normalize_image(image, b="not an integer")
	Traceback (most recent call last):
		...
	AssertionError: b must be a float or an integer
	"""
	# Check input data
	check_image(image, ignore_dtype=ignore_dtype)
	assert isinstance(a, float | int), "a must be a float or an integer"
	assert isinstance(b, float | int), "b must be a float or an integer"
	assert isinstance(method, int), "method must be an integer"
	assert a < b, "a must be less than b"

	# Normalize image
	return cv2.normalize(image, None, a, b, method)




"""
This module provides utility functions for upscaling videos using waifu2x-ncnn-vulkan.

It extracts frames from videos, upscales them individually, and then recombines them
into a high-quality output video. The process preserves audio from the original video
and handles configuration of video encoding parameters like bitrate and framerate.

Main functionalities:

- Extracting frames from videos using FFmpeg
- Upscaling frames using waifu2x-ncnn-vulkan
- Recombining frames into videos with optimized bitrates
- Handling partially processed videos to resume interrupted operations
- Calculating recommended bitrates based on resolution and framerate

The module includes YouTube's recommended bitrate settings for different resolutions,
framerates, and HDR/SDR content types, ensuring optimal quality for various outputs.

Example usage:
.. code-block:: python

	# Imports
	import stouputils.applications.upscaler as app
	from stouputils.io import get_root_path

	# Constants
	ROOT: str = get_root_path(__file__) + "/upscaler"
	INPUT_FOLDER: str = f"{ROOT}/input"
	PROGRESS_FOLDER: str = f"{ROOT}/progress"
	OUTPUT_FOLDER: str = f"{ROOT}/output"

	# Main
	if __name__ == "__main__":
		app.video_upscaler_cli(INPUT_FOLDER, PROGRESS_FOLDER, OUTPUT_FOLDER)
"""

# Imports
import os
import shutil
import subprocess
import sys
from typing import Literal

from PIL import Image

from ...installer import check_executable
from ...io import clean_path
from ...parallel import colored_for_loop, multithreading
from ...print import debug, error, info, warning
from .config import FFMPEG_RELEASES, YOUTUBE_BITRATE_RECOMMENDATIONS, Config
from .image import convert_frame, get_all_files, upscale_folder


# Functions
def get_recommended_bitrate(
	resolution: tuple[int, int], frame_rate: int = 60, upload_type: Literal["SDR","HDR"] = "SDR"
) -> int:
	""" Get the recommended bitrate (in kbps) for the output video based on the video resolution.

	Args:
		resolution  (tuple[int, int]):       Video resolution (width, height).
		frame_rate  (int):                   Frame rate of the video, default is 60.
		upload_type (Literal["SDR","HDR"]):  Upload type from which the recommendation is made, default is "SDR".

	Returns:
		int:     The recommended bitrate for the output video (in kbps)

	Source: https://support.google.com/youtube/answer/1722171?hl=en#zippy=%2Cbitrate

	Examples:
		>>> # Valid examples
		>>> get_recommended_bitrate((3840, 2160), 60, "SDR")
		68000
		>>> get_recommended_bitrate((1920, 1080), 60, "HDR")
		15000
		>>> get_recommended_bitrate((1920, 1080), 60, "SDR")
		12000
		>>> get_recommended_bitrate((1920, 1080), 30, "SDR")
		8000

		>>> # Invalid examples
		>>> get_recommended_bitrate((1920, 1080), 60, "Ratio")
		Traceback (most recent call last):
			...
		AssertionError: Invalid upload type: 'Ratio'
		>>> get_recommended_bitrate("1920x1080", 60, "SDR")
		Traceback (most recent call last):
			...
		AssertionError: Invalid resolution: 1920x1080, must be a tuple of two integers
		>>> get_recommended_bitrate((1920, 1080), -10, "SDR")
		Traceback (most recent call last):
			...
		AssertionError: Invalid frame rate: -10, must be a positive integer
	"""
	# Assertions
	assert isinstance(resolution, tuple) and len(resolution) == 2, \
		f"Invalid resolution: {resolution}, must be a tuple of two integers"
	assert isinstance(frame_rate, int) and frame_rate > 0, \
		f"Invalid frame rate: {frame_rate}, must be a positive integer"
	assert upload_type in YOUTUBE_BITRATE_RECOMMENDATIONS, \
		f"Invalid upload type: '{upload_type}'"

	# Determine frame rate category
	frame_rate_category: str = "high" if frame_rate >= 48 else "standard"

	# Get the appropriate bitrate dictionary
	resolution_bitrates: dict[int, int] = YOUTUBE_BITRATE_RECOMMENDATIONS[upload_type][frame_rate_category]

	# Find the appropriate bitrate based on resolution
	max_dimension: int = min(*resolution)
	for min_resolution, bitrate in sorted(resolution_bitrates.items(), reverse=True):
		if max_dimension >= min_resolution:
			return bitrate

	# Fallback (should never reach here due to the '0' key in dictionaries)
	return 1000


def check_ffmpeg_executable() -> None:
	if not Config.ffmpeg_executable_checked:
		check_executable(Config.FFMPEG_EXECUTABLE, Config.FFMPEG_CHECK_HELP_TEXT, FFMPEG_RELEASES, append_to_path="bin")
		Config.ffmpeg_executable_checked = True

# Routine to handle a video file
def upscale_video(video_file: str, input_folder: str, progress_folder: str, output_folder: str) -> None:
	""" Handles a video file. """
	# Prepare paths
	input_path: str = f"{input_folder}/{video_file}"
	progress_path: str = f"{progress_folder}/{video_file}"
	p_extracted_path: str = f"{progress_path}/extracted"
	p_upscaled_path: str = f"{progress_path}/upscaled"
	output_path: str = f"{output_folder}/{video_file}"
	os.makedirs(p_extracted_path, exist_ok = True)
	os.makedirs(p_upscaled_path, exist_ok = True)

	# Check if executable is installed
	check_ffmpeg_executable()

	## Step 1: Check if the video file is already upscaled or partially processed, if not, extract frames
	# If the video file is already upscaled, skip it
	if os.path.exists(output_path):
		warning(f"'{video_file}' has already been processed, remove it from the output folder to reprocess it.")
		return

	# If the video is already in the list of videos that have been partially processed, ask to restart or skip
	is_partially_processed: bool = len(os.listdir(p_extracted_path)) > 0 and len(os.listdir(p_upscaled_path)) > 0
	if is_partially_processed:
		info(f"'{video_file}' has already been partially processed, do you want to resume the process? (Y/n)")
		if input().lower() == "n":
			shutil.rmtree(p_upscaled_path, ignore_errors = True)
			os.makedirs(p_upscaled_path, exist_ok = True)
			is_partially_processed = False

	# If the video is not partially processed, extract frames
	if not is_partially_processed:
		debug(f"Extracting frames from '{video_file}'...")

		# Extract frames using ffmpeg
		command: list[str] = [Config.FFMPEG_EXECUTABLE, "-i", input_path, f"{p_extracted_path}/%09d.png"]
		subprocess.run(command, capture_output = True)

	# Convert all frames to JPG format
	frames_to_convert: list[str] = get_all_files(p_extracted_path, ".png")
	if frames_to_convert:
		multithreading(convert_frame, frames_to_convert, desc="Converting frames to JPG format")


	## Step 2: Upscale the frames
	# Get all the frames in the progress folder
	all_frames: list[str] = get_all_files(p_extracted_path, ".jpg")
	upscaled_frames: list[str] = get_all_files(p_upscaled_path, ".jpg")

	# If there are frames to upscale,
	if len(all_frames) > len(upscaled_frames):

		# Try to get upscaling ratio if any
		upscale_ratio: int = 2
		if upscaled_frames:
			with Image.open(upscaled_frames[0]) as img:
				upscaled_size: tuple[int, int] = img.size
			with Image.open(all_frames[0]) as img:
				extracted_size: tuple[int, int] = img.size
			upscale_ratio = upscaled_size[0] // extracted_size[0]
			info(f"Detected upscaling ratio: {upscale_ratio}")
		else:
			if "--upscale" in sys.argv:
				upscale_index: int = sys.argv.index("--upscale")
				if upscale_index + 1 < len(sys.argv):
					upscale_ratio = int(sys.argv[upscale_index + 1])
				else:
					error(
						"No upscaling ratio provided with --upscale flag. "
						"Please provide a ratio after the flag. (1/2/4/8/16/32)"
					)
			else:
				info("No upscaling ratio provided, please enter one (1/2/4/8/16/32, default=2):")
				upscale_ratio = int(input() or "2")

		# For each frame that hasn't been upscaled yet, upscale it
		upscale_folder(p_extracted_path, p_upscaled_path, upscale_ratio, slightly_faster_mode=Config.SLIGHTLY_FASTER_MODE)

	## Step 3: Convert the upscaled frames to a video
	# Get the video bitrate
	if Config.VIDEO_FINAL_BITRATE == -1:
		upscaled_frame: str = get_all_files(p_upscaled_path, ".jpg")[0]
		with Image.open(upscaled_frame) as img:
			upscaled_size: tuple[int, int] = img.size
		video_bitrate: int = get_recommended_bitrate(upscaled_size)
	else:
		video_bitrate: int = Config.VIDEO_FINAL_BITRATE

	# Get the framerate of the original video
	original_framerate: str = "60"
	ffprobe_command: list[str] = [
		Config.FFPROBE_EXECUTABLE,                    # Path to the ffprobe executable
		"-v", "error",                                # Set verbosity level to error (only show errors)
		"-select_streams", "v:0",                     # Select the first video stream
		"-show_entries", "stream=r_frame_rate",       # Show only the frame rate information
		"-of", "default=noprint_wrappers=1:nokey=1",  # Format output without wrappers and keys
		input_path,                                   # Path to the input video file
	]
	try:
		result = subprocess.run(ffprobe_command, capture_output=True, text=True, check=True)
		framerate: str = result.stdout.strip()
		if not framerate or '/' not in framerate: # Basic validation
			warning(f"Could not reliably determine framerate for '{video_file}'. Falling back to 60.")
			original_framerate = "60"
		else:
			debug(f"Detected original framerate: {framerate}")
			original_framerate = framerate
	except (subprocess.CalledProcessError, FileNotFoundError) as e:
		warning(f"Failed to get framerate using ffprobe for '{video_file}': {e}. Falling back to 60.")


	# Prepare the command to convert the upscaled frames to a video
	subprocess.run([
		Config.FFMPEG_EXECUTABLE,
		"-framerate", original_framerate,    # Use the original video's framerate for input frames
		"-i", f"{p_upscaled_path}/%09d.jpg", # Use p_upscaled_path, not upscaled_path
		"-i", input_path,                    # Input video for sound and metadata
		"-b:v", f"{video_bitrate}k",         # Set the video bitrate (in kbps)
		*Config.FFMPEG_ARGS,			 	 # Additional arguments from the config
		"-r", original_framerate,            # Set the *output* video framerate
		output_path,                         # Output video
	])


def video_upscaler_cli(input_folder: str, progress_folder: str, output_folder: str) -> None:
	""" Upscales videos from an input folder and saves them to an output folder.

	Uses intermediate folders for extracted and upscaled frames within the progress folder.
	**Handles resuming partially processed videos.**

	Args:
		input_folder    (str): Path to the folder containing input videos.
		progress_folder (str): Path to the folder for storing intermediate files (frames).
		output_folder   (str): Path to the folder where upscaled videos will be saved.
	"""
	# Clean paths
	input_folder = clean_path(input_folder)
	progress_folder = clean_path(progress_folder)
	output_folder = clean_path(output_folder)
	os.makedirs(input_folder, exist_ok = True)
	os.makedirs(progress_folder, exist_ok = True)
	os.makedirs(output_folder, exist_ok = True)

	# Ask if we should shutdown the computer after the script is finished
	info("Do you want to shutdown the computer after the script is finished? (y/N)")
	shutdown_after_script: bool = input().lower() == "y"

	# Collect all video files in the input folder
	videos: list[str] = [file for file in os.listdir(input_folder) if not file.endswith(".md")]

	# Handle each video file
	for video in colored_for_loop(videos, desc="Upscaling videos"):
		upscale_video(video, input_folder, progress_folder, output_folder)

	# Shutdown the computer after the script is finished
	if shutdown_after_script:
		info("Shutting down the computer...")
		if os.name == "nt":
			subprocess.run(["shutdown", "/s", "/t", "0", "/f"], capture_output = False)
		else:
			subprocess.run(["shutdown", "now"], capture_output = False)





# pyright: reportMissingTypeStubs=false

# Imports
from collections.abc import Callable
from typing import Any

from keras.callbacks import Callback
from keras.models import Model
from keras.optimizers import Optimizer


class ProgressiveUnfreezing(Callback):
	""" Callback inspired by the Learning Rate Finder to progressively unfreeze model layers during training.

	Warning: This callback is not compatible with model.fit() as it modifies the trainable state of the model.
	Prefer doing your own training loop instead.

	This callback can operate in two modes:
	1. Start with all layers frozen and incrementally unfreeze them from 0% to 100% (progressive_freeze=False)
	2. Start with all layers unfrozen and incrementally freeze them from 100% to 0% (progressive_freeze=True)
	"""

	def __init__(
		self,
		base_model: Model,
		steps_per_epoch: int,
		epochs: int,
		reset_weights: bool = False,
		reset_optimizer_function: Callable[[], Optimizer] | None = None,
		update_per_epoch: bool = True,
		update_interval: int = 5,
		progressive_freeze: bool = False
	) -> None:
		""" Initialize the progressive unfreezing callback.

		Args:
			base_model         (Model):   Base model to unfreeze.
			steps_per_epoch    (int):     Number of steps per epoch.
			epochs             (int):     Total number of epochs.
			reset_weights      (bool):    If True, reset weights after each unfreeze.
			reset_optimizer_function (Callable | None):
				If set, use this function to reset the optimizer every update_interval.
				The function should return a compiled optimizer, ex: `lambda: model._get_optimizer(AdamW(...))`.
			update_per_epoch   (bool):    If True, unfreeze per epoch, else per batch.
			update_interval    (int):     Number of steps between each unfreeze to allow model to stabilize.
			progressive_freeze (bool):    If True, start with all layers unfrozen and progressively freeze them.
		"""
		super().__init__()
		self.base_model: Model = base_model
		""" Base model to unfreeze. """
		self.model: Model
		""" Model to apply the progressive unfreezing to. """
		self.steps_per_epoch: int = int(steps_per_epoch)
		""" Number of steps per epoch. """
		self.epochs: int = int(epochs)
		""" Total number of epochs. """
		self.reset_weights: bool = bool(reset_weights)
		""" If True, reset weights after each unfreeze. """
		self.reset_optimizer_function: Callable[[], Optimizer] | None = reset_optimizer_function
		""" If reset_weights is True and this is not None, use this function to get a new optimizer. """
		self.update_per_epoch: bool = bool(update_per_epoch)
		""" If True, unfreeze per epoch, else per batch. """
		self.update_interval: int = max(1, int(update_interval))
		""" Number of steps between each unfreeze to allow model to stabilize. """
		self.progressive_freeze: bool = bool(progressive_freeze)
		""" If True, start with all layers unfrozen and progressively freeze them. """

		# If updating per epoch, remove to self.epochs the update interval to allow the last step to train with 100% unfreeze
		if self.update_per_epoch:
			self.epochs -= self.update_interval

		# Calculate total steps considering the update interval
		total_steps_raw: int = self.epochs if self.update_per_epoch else self.steps_per_epoch * self.epochs
		self.total_steps: int = total_steps_raw // self.update_interval
		""" Total number of update steps (considering update_interval). """

		self.fraction_unfrozen: list[float] = []
		""" Fraction of layers unfrozen. """
		self.losses: list[float] = []
		""" Losses. """
		self._all_layers: list[Any] = []
		""" All layers. """
		self._initial_trainable: list[bool] = []
		""" Initial trainable states. """
		self._initial_weights: list[Any] | None = None
		""" Initial weights of the model. """
		self._last_update_step: int = -1
		""" Last step when layers were unfrozen. """
		self.params: dict[str, Any]

	def on_train_begin(self, logs: dict[str, Any] | None = None) -> None:
		""" Set initial layer trainable states at the start of training and store initial states and weights.

		Args:
			logs (dict | None): Training logs.
		"""
		# Collect all layers from the model and preserve their original trainable states for potential restoration
		self._all_layers = self.base_model.layers
		self._initial_trainable = [bool(layer.trainable) for layer in self._all_layers]

		# Store initial weights to reset after each unfreeze
		if self.reset_weights:
			self._initial_weights = self.model.get_weights()

		# Set initial trainable state based on mode
		for layer in self._all_layers:
			layer.trainable = self.progressive_freeze  # If progressive_freeze, start with all layers unfrozen

	def _update_layers(self, step: int) -> None:
		""" Update layer trainable states based on the current step and mode.
		Reset weights after each update to prevent bias in the results.

		Args:
			step (int): Current training step.
		"""
		# Calculate the effective step considering the update interval
		effective_step: int = step // self.update_interval

		# Skip if we haven't reached the next update interval
		if effective_step <= self._last_update_step:
			return
		self._last_update_step = effective_step

		# Calculate the number of layers to unfreeze based on current effective step
		n_layers: int = len(self._all_layers)

		if self.progressive_freeze:
			# For progressive freezing, start at 1.0 (all unfrozen) and decrease to 0.0
			fraction: float = max(0.0, 1.0 - (effective_step + 1) / self.total_steps)
		else:
			# For progressive unfreezing, start at 0.0 (all frozen) and increase to 1.0
			fraction: float = min(1.0, (effective_step + 1) / self.total_steps)

		n_unfreeze: int = int(n_layers * fraction)  # Number of layers to keep unfrozen
		self.fraction_unfrozen.append(fraction)

		# Set trainable state for each layer based on position
		# For both modes, we unfreeze from the top (output layers) to the bottom (input layers)
		for i, layer in enumerate(self._all_layers):
			layer.trainable = i >= (n_layers - n_unfreeze)

		# Reset weights to initial state to prevent bias and reset optimizer
		if self._initial_weights is not None:
			self.model.set_weights(self._initial_weights) # pyright: ignore [reportUnknownMemberType]
		if self.reset_optimizer_function is not None:
			self.model.optimizer = self.reset_optimizer_function()
			self.model.optimizer.build(self.model.trainable_variables) # pyright: ignore [reportUnknownMemberType]

	def _track_loss(self, logs: dict[str, Any] | None = None) -> None:
		""" Track the current loss.

		Args:
			logs (dict | None): Training logs containing loss information.
		"""
		if logs and "loss" in logs:
			self.losses.append(logs["loss"])

	def on_batch_begin(self, batch: int, logs: dict[str, Any] | None = None) -> None:
		""" Update layer trainable states at the start of each batch if not updating per epoch.

		Args:
			batch (int): Current batch index.
			logs (dict | None): Training logs.
		"""
		# Skip if we're updating per epoch instead of per batch
		if self.update_per_epoch:
			return

		# Calculate the current step across all epochs and update layers
		step: int = self.params.get("steps", self.steps_per_epoch) * self.params.get("epoch", 0) + batch
		self._update_layers(step)

	def on_batch_end(self, batch: int, logs: dict[str, Any] | None = None) -> None:
		""" Track loss at the end of each batch if not updating per epoch.

		Args:
			batch (int): Current batch index.
			logs (dict | None): Training logs.
		"""
		# Skip if we're updating per epoch instead of per batch
		if self.update_per_epoch:
			return

		# Record the loss if update interval is reached
		if batch % self.update_interval == 0:
			self._track_loss(logs)

	def on_epoch_begin(self, epoch: int, logs: dict[str, Any] | None = None) -> None:
		""" Update layer trainable states at the start of each epoch if updating per epoch.

		Args:
			epoch (int): Current epoch index.
			logs (dict | None): Training logs.
		"""
		# Skip if we're updating per batch instead of per epoch
		if not self.update_per_epoch:
			return

		# Update layers based on current epoch
		self._update_layers(epoch)

	def on_epoch_end(self, epoch: int, logs: dict[str, Any] | None = None) -> None:
		""" Track loss at the end of each epoch if updating per epoch.

		Args:
			epoch (int): Current epoch index.
			logs (dict | None): Training logs.
		"""
		# Skip if we're updating per batch instead of per epoch
		if not self.update_per_epoch:
			return

		# Record the loss if update interval is reached
		if epoch % self.update_interval == 0:
			self._track_loss(logs)

	def on_train_end(self, logs: dict[str, Any] | None = None) -> None:
		""" Restore original trainable states at the end of training.

		Args:
			logs (dict | None): Training logs.
		"""
		# Restore each layer's original trainable state
		for layer, trainable in zip(self._all_layers, self._initial_trainable, strict=False):
			layer.trainable = trainable

	def get_results(self, multiply_by_100: bool = True) -> tuple[list[float], list[float]]:
		""" Get the results of the progressive unfreezing from 0% to 100% even if progressive_freeze is True.

		Args:
			multiply_by_100 (bool): If True, multiply the fractions by 100 to get percentages.

		Returns:
			tuple[list[float], list[float]]: fractions of layers unfrozen, and losses.
		"""
		fractions: list[float] = self.fraction_unfrozen

		# Reverse the order if progressive_freeze is True
		if self.progressive_freeze:
			fractions = fractions[::-1]

		# Multiply by 100 if requested
		if multiply_by_100:
			fractions = [x * 100 for x in fractions]

		# Return the results
		return fractions, self.losses




"""
This module contains the RangeTuple class, which provides a named tuple for range parameters.

This class contains methods for:

- Iterating over range values
- Accessing range values by index
- Slicing range values
- Converting to string representation
"""
# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false
# pyright: reportIncompatibleMethodOverride=false

# Imports
from __future__ import annotations

from collections.abc import Generator
from typing import NamedTuple

import numpy as np

from .utils import Utils


# Create base tuple class
class _RangeTupleBase(NamedTuple):
	""" Base class for RangeTuple """
	mini: float | None
	""" The minimum value (inclusive) (can be None if default is set) """
	maxi: float | None
	""" The maximum value (exclusive) (can be None if default is set) """
	step: float | None
	""" The step value between elements (can be None if default is set) """
	default: float | None
	""" Optional default value, usually middle of range """


# Tuple class for range parameters
class RangeTuple(_RangeTupleBase):
	""" A named tuple containing range parameters.

	Attributes:
		mini     (float):        The minimum value (inclusive) (can be None if default is set)
		maxi     (float):        The maximum value (exclusive) (can be None if default is set)
		step     (float):        The step value between elements (can be None if default is set)
		default  (float|None):   Optional default value, usually middle of range

	Examples:
		>>> r = RangeTuple(mini=0.0, maxi=1.0, step=0.3)
		>>> print(r)
		mini=0.0, maxi=1.0, step=0.3, default=None
		>>> [int(x*10) for x in r]
		[0, 3, 6, 9]
		>>> len(r)
		4
		>>> r[0]
		0.0
		>>> r[100], r[99]	# High indexes will bypass the maximum value
		(30.0, 29.7)
		>>> r[1:3]
		[0.3, 0.6]
		>>> round(r[-2], 1)
		0.6
		>>> r = RangeTuple()
		Traceback (most recent call last):
			...
		ValueError: RangeTuple parameters must not be None
	"""
	def __new__(
		cls,
		mini: float | None = None,
		maxi: float | None = None,
		step: float | None = 1.0,
		default: float | None = None
	) -> RangeTuple:
		if (mini is None or maxi is None):
			if default is None:
				raise ValueError("RangeTuple parameters must not be None")
			else:
				step = None
		return super().__new__(cls, mini, maxi, step, default)

	def __str__(self) -> str:
		return f"mini={self.mini}, maxi={self.maxi}, step={self.step}, default={self.default}"

	def __repr__(self) -> str:
		return f"RangeTuple(mini={self.mini!r}, maxi={self.maxi!r}, step={self.step!r}, default={self.default!r})"

	def __iter__(self) -> Generator[float, None, None]:
		""" Iterate over the range values.
		If the range is not initialized (mini or maxi is None), yield the default value.
		Else, yield from np.arange(...)

		Returns:
			Iterator[float]: Iterator over the range values

		Examples:
			>>> r = RangeTuple(mini=0.0, maxi=1.0, step=0.5)
			>>> list(r)
			[0.0, 0.5]
			>>> r = RangeTuple(default=1.0)
			>>> list(r)
			[1.0]
		"""
		if self.mini is None or self.maxi is None or self.step is None:
			yield float(self.default) # pyright: ignore [reportArgumentType]
		else:
			yield from np.arange(self.mini, self.maxi, self.step)

	def __len__(self) -> int:
		""" Return the number of values in the range.

		Returns:
			int: Number of values in the range

		Examples:
			>>> len(RangeTuple(mini=0.0, maxi=1.0, step=0.5))
			3
			>>> len(RangeTuple(default=1.0))
			1
		"""
		if self.mini is None or self.maxi is None or self.step is None:
			return 1
		else:
			return int((self.maxi - self.mini) / self.step) + 1

	def __getitem__(self, index: int | slice) -> float | list[float]:
		""" Get value(s) at the given index or slice.
		If the range is not initialized, return the default value.

		Args:
			index (int | slice): Index or slice to get values for
		Returns:
			float | list[float]: Value(s) at the specified index/slice

		Examples:
			>>> r = RangeTuple(mini=0.0, maxi=1.0, step=0.5)
			>>> r[0]
			0.0
			>>> r[1]
			0.5
			>>> r[-1]
			1.0
			>>> r[0:2]
			[0.0, 0.5]
			>>> r = RangeTuple(default=1.0)
			>>> r[0]
			1.0
			>>> r[1]
			1.0
		"""
		if self.mini is None or self.maxi is None or self.step is None:
			if self.default is not None:
				return self.default
			else:
				raise ValueError("RangeTuple is not initialized")
		else:
			if isinstance(index, slice):
				# Handle None values in slice by using defaults
				start: int = 0 if index.start is None else index.start
				stop: int = len(self) if index.stop is None else index.stop
				step: int = 1 if index.step is None else index.step

				return [self.mini + i * self.step for i in range(start, stop, step)]
			else:
				while index < 0:
					index = len(self) + index
				return float(self.mini + index * self.step)

	def __mul__(self, other: float) -> RangeTuple:
		""" Multiply the range by a factor.

		Args:
			other (float): Factor to multiply by
		Returns:
			RangeTuple: New range with all values multiplied by the factor

		Examples:
			>>> r = RangeTuple(mini=1.0, maxi=2.0, step=0.5)
			>>> r * 2
			RangeTuple(mini=2.0, maxi=4.0, step=1.0, default=None)
			>>> r = RangeTuple(default=1.0)
			>>> r * 3
			RangeTuple(mini=None, maxi=None, step=None, default=3.0)
		"""
		return RangeTuple(
			mini=Utils.safe_multiply_none(self.mini, other),
			maxi=Utils.safe_multiply_none(self.maxi, other),
			step=Utils.safe_multiply_none(self.step, other),
			default=Utils.safe_multiply_none(self.default, other)
		)

	def __truediv__(self, other: float) -> RangeTuple:
		""" Divide the range by a factor.

		Args:
			other (float): Factor to divide by
		Returns:
			RangeTuple: New range with all values divided by the factor

		Examples:
			>>> r = RangeTuple(mini=2.0, maxi=4.0, step=1.0)
			>>> r / 2
			RangeTuple(mini=1.0, maxi=2.0, step=0.5, default=None)
			>>> r = RangeTuple(default=6.0)
			>>> r / 3
			RangeTuple(mini=None, maxi=None, step=None, default=2.0)
		"""
		return RangeTuple(
			mini=Utils.safe_divide_none(self.mini, other),
			maxi=Utils.safe_divide_none(self.maxi, other),
			step=Utils.safe_divide_none(self.step, other),
			default=Utils.safe_divide_none(self.default, other)
		)

	def random(self) -> float:
		""" Return a random value from the range.
		If the range is not initialized, return the default value.

		Returns:
			float: Random value from the range

		Examples:
			>>> r = RangeTuple(mini=0.0, maxi=1.0, step=1.0)
			>>> 0.0 <= r.random() <= 1.0
			True
			>>> r = RangeTuple(default=1.0)
			>>> r.random()
			1.0
		"""
		index = np.random.randint(0, len(self))
		return self.__getitem__(index) # pyright: ignore [reportReturnType]





# Imports
from .common import Any, NDArray, check_image, np


# Function
def invert_image(image: NDArray[Any], ignore_dtype: bool = False) -> NDArray[Any]:
	""" Invert the colors of an image.

	This function inverts the colors of the input image by subtracting each pixel value
	from the maximum possible value based on the image type.

	Args:
		image         (NDArray[Any]):       Input image as a NumPy array.
		ignore_dtype  (bool):             Ignore the dtype check.
	Returns:
		NDArray[Any]: Image with inverted colors.

	>>> ## Basic tests
	>>> image = np.array([[10, 20, 30], [40, 50, 60], [70, 80, 90]], dtype=np.uint8)
	>>> inverted = invert_image(image)
	>>> inverted.tolist()
	[[245, 235, 225], [215, 205, 195], [185, 175, 165]]

	>>> # Test with floating point image
	>>> float_img = np.array([[0.1, 0.2], [0.3, 0.4]], dtype=np.float32)
	>>> [round(x, 1) for x in invert_image(float_img).flatten()]
	[0.9, 0.8, 0.7, 0.6]

	>>> # Test with RGB image
	>>> rgb = np.zeros((2, 2, 3), dtype=np.uint8)
	>>> rgb[0, 0] = [255, 0, 0]  # Red pixel
	>>> inverted_rgb = invert_image(rgb)
	>>> inverted_rgb[0, 0].tolist()
	[0, 255, 255]

	>>> ## Test invalid inputs
	>>> invert_image("not an image")
	Traceback (most recent call last):
		...
	AssertionError: Image must be a numpy array
		"""
	# Check input data
	check_image(image, ignore_dtype=ignore_dtype)

	# Get the maximum value based on the image's data type
	if image.dtype == np.uint8:
		max_value = 255
	elif image.dtype == np.uint16:
		max_value = 65535
	elif image.dtype == np.float32 or image.dtype == np.float64:
		# For float images, we assume range [0, 1]
		max_value = 1.0
	else:
		# Default case, assuming 8-bit
		max_value = 255
		image = image.astype(np.uint8)

	# Invert the image
	inverted = max_value - image

	# Ensure we return the same dtype as the input
	return inverted.astype(image.dtype)




""" Load configuration from the set.py file and handle some special cases.

Proper way to get the configuration is by importing this module, not the set.py file directly.
"""

# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false
# pyright: reportMissingTypeStubs=false

# Imports
import os
from typing import Any

from .set import DataScienceConfig

# Special cases
# Hide GPU when using CPU
if DataScienceConfig.TENSORFLOW_DEVICE.lower().startswith("/cpu"):
	os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

# Precise which GPU we use
elif DataScienceConfig.TENSORFLOW_DEVICE.lower().startswith("/gpu"):
	os.environ["CUDA_VISIBLE_DEVICES"] = DataScienceConfig.TENSORFLOW_DEVICE.split(":")[-1]

	# Configure TensorFlow (if available)
	try:
		from tensorflow import config as tf_config

		# Get the physical devices
		physical_devices: list[Any] = tf_config.list_physical_devices("GPU")

		# Configure TensorFlow GPU memory management to allocate memory dynamically
		# This prevents TensorFlow from allocating all GPU memory upfront
		# Instead, memory will grow as needed, allowing better resource sharing
		for device in physical_devices:
			tf_config.experimental.set_memory_growth(device, True)

		# Disable eager execution mode in TensorFlow
		# This improves performance by allowing TensorFlow to create an optimized graph
		# of operations instead of executing operations one by one (at the cost of debugging difficulty)
		tf_config.run_functions_eagerly(False)
	except ImportError:
		pass

	# Enable mixed precision training (if available)
	try:
		from keras import mixed_precision
		mixed_precision.set_global_policy(DataScienceConfig.MIXED_PRECISION_POLICY)
	except ImportError:
		pass





# pyright: reportUnusedImport=false
# ruff: noqa: F401

# Imports
from .common import Any, NDArray, check_image, cv2, np


# Functions
def sharpen_image(image: NDArray[Any], alpha: float, ignore_dtype: bool = False) -> NDArray[Any]:
	""" Sharpen an image.

	Args:
		image         (NDArray[Any]):  Image to sharpen
		alpha         (float):       Sharpening factor
		ignore_dtype  (bool):        Ignore the dtype check
	Returns:
		NDArray[Any]: Sharpened image

	>>> ## Basic tests
	>>> image = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
	>>> sharpened = sharpen_image(image.astype(np.uint8), 1.5)
	>>> sharpened.shape == image.shape
	True

	>>> img = np.full((5,5), 128, dtype=np.uint8)
	>>> img[2,2] = 255  # Center bright pixel
	>>> sharp = sharpen_image(img, 1.0)
	>>> sharp[2,2] > img[2,2] * 0.9  # Center should stay bright
	True

	>>> rgb = np.full((3,3,3), 128, dtype=np.uint8)
	>>> sharp_rgb = sharpen_image(rgb, 1.0)
	>>> sharp_rgb.shape == (3,3,3)
	True

	>>> ## Test invalid inputs
	>>> sharpen_image("not an image", 1.5)
	Traceback (most recent call last):
		...
	AssertionError: Image must be a numpy array

	>>> sharpen_image(image.astype(np.uint8), "1.5")
	Traceback (most recent call last):
		...
	AssertionError: alpha must be a number, got <class 'str'>
	"""
	# Check input data
	check_image(image, ignore_dtype=ignore_dtype)
	assert isinstance(alpha, float | int), f"alpha must be a number, got {type(alpha)}"

	# Apply sharpening
	blurred: NDArray[Any] = cv2.GaussianBlur(image, (0, 0), 3)
	return cv2.addWeighted(image, 1 + alpha, blurred, -alpha, 0)





# Imports
from PIL import Image

from .common import Any, NDArray, check_image, np


# Functions
def resize_image(
	image: NDArray[Any],
	width: int,
	height: int,
	resample: Image.Resampling | int = Image.Resampling.LANCZOS,
	ignore_dtype: bool = False
) -> NDArray[Any]:
	""" Resize an image to a new width and height.

	Args:
		image         (NDArray[Any]):              Image to resize
		width         (int):                     New width
		height        (int):                     New height
		resample      (Image.Resampling | int):  Resampling method
		ignore_dtype  (bool):                    Ignore the dtype check
	Returns:
		NDArray[Any]: Image with resized dimensions

	>>> ## Basic tests
	>>> image = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
	>>> resized = resize_image(image.astype(np.uint8), 6, 6)
	>>> resized.shape
	(6, 6)

	>>> img = np.ones((5, 5), dtype=np.uint8) * 255
	>>> resized = resize_image(img, 10, 10)
	>>> resized.shape
	(10, 10)

	>>> rgb = np.full((3, 3, 3), 128, dtype=np.uint8)
	>>> resized_rgb = resize_image(rgb, 6, 6)
	>>> resized_rgb.shape
	(6, 6, 3)

	>>> ## Test invalid inputs
	>>> resize_image("not an image", 10, 10)
	Traceback (most recent call last):
		...
	AssertionError: Image must be a numpy array

	>>> resize_image(image.astype(np.uint8), "10", 10)
	Traceback (most recent call last):
		...
	AssertionError: width must be integer, got <class 'str'>

	>>> resize_image(image.astype(np.uint8), 10, "10")
	Traceback (most recent call last):
		...
	AssertionError: height must be integer, got <class 'str'>
	"""
	# Check input data
	check_image(image, ignore_dtype=ignore_dtype)
	assert isinstance(width, int), f"width must be integer, got {type(width)}"
	assert isinstance(height, int), f"height must be integer, got {type(height)}"
	assert isinstance(resample, Image.Resampling | int), f"resample must be Image.Resampling, got {type(resample)}"

	# Resize image
	new_image: Image.Image = Image.fromarray(image)
	new_image = new_image.resize((width, height), resample=resample)
	return np.array(new_image)




"""
This module provides functions for creating and managing archives.

- make_archive: Create a zip archive from a source directory with consistent file timestamps.
- repair_zip_file: Try to repair a corrupted zip file by ignoring some of the errors

.. image:: https://raw.githubusercontent.com/Stoupy51/stouputils/refs/heads/main/assets/archive_module.gif
  :alt: stouputils archive examples
"""

# Imports
import os
from zipfile import ZIP_DEFLATED, ZipFile, ZipInfo

from .decorators import LogLevels, handle_error
from .dont_look.zip_file_override import ZipFileOverride
from .io import clean_path, super_copy


# Function that makes an archive with consistency (same zip file each time)
@handle_error()
def make_archive(
	source: str,
	destinations: list[str] | str | None = None,
	override_time: None | tuple[int, int, int, int, int, int] = None,
	create_dir: bool = False
) -> bool:
	""" Create a zip archive from a source directory with consistent file timestamps.
	(Meaning deterministic zip file each time)

	Creates a zip archive from the source directory and copies it to one or more destinations.
	The archive will have consistent file timestamps across runs if override_time is specified.
	Uses maximum compression level (9) with ZIP_DEFLATED algorithm.

	Args:
		source				(str):						The source folder to archive
		destinations		(list[str]|str):			The destination folder(s) or file(s) to copy the archive to
		override_time		(None | tuple[int, ...]):	The constant time to use for the archive
			(e.g. (2024, 1, 1, 0, 0, 0) for 2024-01-01 00:00:00)
		create_dir			(bool):						Whether to create the destination directory if it doesn't exist
	Returns:
		bool: Always returns True unless any strong error
	Examples:

	.. code-block:: python

		> make_archive("/path/to/source", "/path/to/destination.zip")
		> make_archive("/path/to/source", ["/path/to/destination.zip", "/path/to/destination2.zip"])
		> make_archive("src", "hello_from_year_2085.zip", override_time=(2085,1,1,0,0,0))
	"""
	# Fix copy_destinations type if needed
	if destinations is None:
		destinations = []
	if destinations and isinstance(destinations, str):
		destinations = [destinations]
	if not destinations:
		raise ValueError("destinations must be a list of at least one destination")

	# Create the archive
	destination: str = clean_path(destinations[0])
	destination = destination if ".zip" in destination else destination + ".zip"
	with ZipFile(destination, "w", compression=ZIP_DEFLATED, compresslevel=9) as zip:
		for root, _, files in os.walk(source):
			for file in files:
				file_path: str = clean_path(os.path.join(root, file))
				info: ZipInfo = ZipInfo(file_path)
				info.compress_type = ZIP_DEFLATED
				if override_time:
					info.date_time = override_time
				with open(file_path, "rb") as f:
					zip.writestr(info, f.read())

	# Copy the archive to the destination(s)
	for dest_folder in destinations[1:]:
		@handle_error(Exception, message=f"Unable to copy '{destination}' to '{dest_folder}'", error_log=LogLevels.WARNING)
		def internal(src: str, dest: str) -> None:
			super_copy(src, dest, create_dir=create_dir)
		internal(destination, clean_path(dest_folder))

	return True



# Function that repair a corrupted zip file (ignoring some of the errors)
@handle_error()
def repair_zip_file(file_path: str, destination: str) -> bool:
	""" Try to repair a corrupted zip file by ignoring some of the errors

	Args:
		file_path		(str):	Path of the zip file to repair
		destination		(str):	Destination of the new file
	Returns:
		bool: Always returns True unless any strong error

	Examples:

	.. code-block:: python

		> repair_zip_file("/path/to/source.zip", "/path/to/destination.zip")
	"""
	# Check
	if not os.path.exists(file_path):
		raise FileNotFoundError(f"File '{file_path}' not found")
	dirname: str = os.path.dirname(destination)
	if dirname and not os.path.exists(dirname):
		raise FileNotFoundError(f"Directory '{dirname}' not found")

	# Read it
	with ZipFileOverride(file_path, "r") as zip_file:

		# Get a list of all the files in the ZIP file
		file_list: list[str] = zip_file.namelist()

		# Create a new ZIP file at the destination
		with ZipFileOverride(destination, "w", compression=ZIP_DEFLATED) as new_zip_file:
			for file_name in file_list:
				try:
					new_zip_file.writestr(file_name, zip_file.read(file_name))
				except KeyboardInterrupt:
					continue

	return True






# Imports
from .common import Any, NDArray, check_image, cv2, np


# Functions
def threshold_image(image: NDArray[Any], thresholds: list[float], ignore_dtype: bool = False) -> NDArray[Any]:
	""" Apply multi-level threshold to an image.

	Args:
		image         (NDArray[Any]):      Image to threshold
		threshold     (list[float]):     List of threshold values (between 0 and 1)
		ignore_dtype  (bool):            Ignore the dtype check
	Returns:
		NDArray[Any]: Multi-level thresholded image

	>>> ## Basic tests
	>>> image = np.array([[100, 150, 200], [50, 125, 175], [25, 75, 225]])
	>>> threshold_image(image.astype(np.uint8), [0.3, 0.6]).tolist()
	[[85, 85, 170], [0, 85, 170], [0, 0, 170]]

	>>> rgb = np.random.randint(0, 256, (3,3,3), dtype=np.uint8)
	>>> thresh_rgb = threshold_image(rgb, [0.3, 0.6])
	>>> thresh_rgb.shape == rgb.shape
	True

	>>> ## Test invalid inputs
	>>> threshold_image("not an image", [0.5])
	Traceback (most recent call last):
		...
	AssertionError: Image must be a numpy array

	>>> threshold_image(image.astype(np.uint8), [1.5])
	Traceback (most recent call last):
		...
	AssertionError: threshold values must be between 0 and 1, got [1.5]
	"""
	# Check input data
	check_image(image, ignore_dtype=ignore_dtype)
	assert all(0 <= t <= 1 for t in thresholds), f"threshold values must be between 0 and 1, got {thresholds}"

	# Convert thresholds from 0-1 range to 0-255 range
	threshold_values: list[int] = [int(t * 255) for t in sorted(thresholds)]

	# Apply threshold
	if len(image.shape) == 2:
		# Grayscale image
		result: NDArray[Any] = np.zeros_like(image)
		for i, thresh in enumerate(threshold_values):
			value: int = int(255 * (i + 1) / (len(threshold_values) + 1))
			mask: NDArray[Any] = cv2.threshold(image, thresh, value, cv2.THRESH_BINARY)[1]
			result = cv2.max(result, mask)
	else:
		# Color image - convert to grayscale first, then back to color
		gray: NDArray[Any] = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
		result: NDArray[Any] = np.zeros_like(gray)
		for i, thresh in enumerate(threshold_values):
			value: int = int(255 * (i + 1) / (len(threshold_values) + 1))
			mask: NDArray[Any] = cv2.threshold(gray, thresh, value, cv2.THRESH_BINARY)[1]
			result = cv2.max(result, mask)
		result = cv2.cvtColor(result, cv2.COLOR_GRAY2BGR)

	return result





# Imports
import argparse
import itertools
from typing import Any, Literal

from ...decorators import handle_error, measure_time
from ...print import info, error, progress
from ...io import clean_path
from ..config.get import DataScienceConfig
from ..dataset import LOWER_GS, Dataset, DatasetLoader, GroupingStrategy, XyTuple
from ..models.all import ALL_MODELS, CLASS_MAP, ModelInterface

# Constants
MODEL_HELP: str             = "Model(s) name or alias to use"
INPUT_HELP: str             = "Path to the dataset, ex: 'data/aug_hip_implant'"
BASED_OF_HELP: str          = "Path to the base dataset for filtering train/test, ex: 'data/hip_implant'"
TRANSFER_LEARNING_HELP: str = "Transfer learning source (imagenet, None, 'data/dataset_folder')"
GROUPING_HELP: str          = "Grouping strategy for the dataset"
K_FOLD_HELP: str            = "Number of folds for k-fold cross validation (0 = no k-fold, negative = LeavePOut)"
GRID_SEARCH_HELP: str       = "If grid search should be performed on hyperparameters"
VERBOSE_HELP: str           = "Verbosity level sent to functions"
PARSER_DESCRIPTION: str     = "Command-line interface for training and evaluating machine learning models."


# Main function
@measure_time(info, "Total execution time of the script")
@handle_error(exceptions=(KeyboardInterrupt, Exception), error_log=DataScienceConfig.ERROR_LOG)
def routine(
	default_input: str = f"{DataScienceConfig.DATA_FOLDER}/aug_hip_implant_preprocessed",
	default_based_of: str = "auto",
	default_transfer_learning: str = "imagenet",
	default_grouping_strategy: str = "none",
	default_kfold: int = 0,
	default_verbose: int = 100,

	loading_type: Literal["image"] = "image",
	grid_search_param_grid: dict[str, list[Any]] | None = None,
	add_to_train_only: list[str] | None = None,
) -> None:
	""" Main function of the script for training and evaluating machine learning models.

	This function handles the entire workflow for model training and evaluation, including:
	- Parsing command-line arguments (default values are set in the function signature)
	- Loading and preparing datasets with configurable grouping strategies
	- Supporting transfer learning from various sources
	- Enabling K-fold cross-validation, LeavePOut or LeaveOneOut
	- Providing grid search capabilities for hyperparameter optimization
	- Incorporating additional training data from specified paths

	Args:
		default_input             (str):              Default path to the dataset to use.
		default_based_of          (str):              Default path to the base dataset for filtering train/test data.
		default_transfer_learning (str):              Default transfer learning source.
		default_grouping_strategy (str):              Default grouping strategy for the dataset.
		default_kfold             (int):              Default number of folds for k-fold cross validation.
		default_verbose           (int):              Default verbosity level.
		loading_type              (Literal["image"]):             Type of data to load, currently only supports "image".
		grid_search_param_grid    (dict[str, list[Any]] | None):  Parameters grid for hyperparameter optimization.
		add_to_train_only         (list[str] | None):             List of paths to additional training datasets.

	Returns:
		None: This function does not return anything.
	"""
	if grid_search_param_grid is None:
		grid_search_param_grid = {"batch_size": [8, 16, 32, 64]}
	if add_to_train_only is None:
		add_to_train_only = []

	info("Starting the script...")

	# Parse the arguments
	parser = argparse.ArgumentParser(description=PARSER_DESCRIPTION)
	parser.add_argument("--model",             type=str, choices=ALL_MODELS, required=True, help=MODEL_HELP)
	parser.add_argument("--input",             type=str, default=default_input, help=INPUT_HELP)
	parser.add_argument("--based_of",          type=str, default=default_based_of, help=BASED_OF_HELP)
	parser.add_argument("--transfer_learning", type=str, default=default_transfer_learning, help=TRANSFER_LEARNING_HELP)
	parser.add_argument("--grouping_strategy", type=str, default=default_grouping_strategy, choices=LOWER_GS, help=GROUPING_HELP)
	parser.add_argument("--kfold", type=int, default=default_kfold, help=K_FOLD_HELP)
	parser.add_argument("--grid_search", action="store_true", help=GRID_SEARCH_HELP)
	parser.add_argument("--verbose", type=int, default=default_verbose, help=VERBOSE_HELP)
	args: argparse.Namespace = parser.parse_args()
	model: str = args.model.lower()
	kfold: int = args.kfold
	input_path: str = clean_path(args.input, trailing_slash=False)
	based_of: str = clean_path(args.based_of, trailing_slash=False)
	transfer_learning: str = clean_path(args.transfer_learning, trailing_slash=False)
	verbose: int = args.verbose
	grouping_strategy: str = args.grouping_strategy
	grid_search: bool = args.grid_search

	# If based_of is "auto", set it to the input path without the "aug"
	if based_of == "auto":
		prefix: str = "/" + DataScienceConfig.AUGMENTED_DIRECTORY_PREFIX
		if prefix in input_path:
			based_of = input_path.replace(prefix, "/")
		else:
			based_of = ""

	# Load the dataset
	kwargs: dict[str, Any] = {}
	if grouping_strategy == "concatenate":
		kwargs["color_mode"] = "grayscale"
	dataset: Dataset = DatasetLoader.from_path(
		path=input_path,
		loading_type=loading_type,
		seed=DataScienceConfig.SEED,
		test_size=DataScienceConfig.TEST_SIZE,
		grouping_strategy=next(x for x in GroupingStrategy if x.name.lower() == grouping_strategy),
		based_of=based_of,
		**kwargs
	)
	info(dataset)

	# Define parameter combinations
	param_combinations: list[dict[str, Any]] = [{}]  # Default empty params
	if grid_search:

		# Generate all parameter combinations
		param_combinations.clear()
		for values in itertools.product(*grid_search_param_grid.values()):
			param_combinations.append(dict(zip(grid_search_param_grid.keys(), values, strict=False)))

	# Load additional training data from provided paths
	additional_training_data: XyTuple = XyTuple.empty()
	for path in add_to_train_only:
		try:
			additional_dataset: Dataset = DatasetLoader.from_path(
				path=path,
				loading_type=loading_type,
				seed=DataScienceConfig.SEED,
				test_size=0,  # Use all data for training
				**kwargs
			)
			additional_training_data += additional_dataset.training_data
		except Exception as e:
			error(f"Failed to load additional training data from '{path}': {e}")

	# Prepare the initialization arguments
	# (num_classes: int, kfold: int = 0, transfer_learning: str = "imagenet", **override_params: Any)
	initialization_args: dict[str, Any] = {

		# Mandatory arguments
		"num_classes": dataset.num_classes,
		"kfold": kfold,
		"transfer_learning": transfer_learning,

		# Optional arguments (override_params)
		"additional_training_data": additional_training_data
	}

	# Collect all class routines that match the model name
	classes: list[type[ModelInterface]] = [key for key, values in CLASS_MAP.items() if model in values]

	# For each parameter combination
	for i, params in enumerate(param_combinations):
		if grid_search:
			progress(f"Grid search {i+1}/{len(param_combinations)}, Training with parameters:\n{params}")
			initialization_args["override_params"] = params

		# Launch all class routines
		for class_to_process in classes:
			model_instance: ModelInterface = class_to_process(**initialization_args)
			trained_model: ModelInterface = model_instance.routine_full(dataset, verbose)
			info(trained_model)
			del trained_model
	return




""" This module contains utilities for PyPI.

- pypi_full_routine: Upload the most recent file(s) to PyPI after updating pip and required packages and building the package

.. image:: https://raw.githubusercontent.com/Stoupy51/stouputils/refs/heads/main/assets/continuous_delivery/pypi_module.gif
  :alt: stouputils pypi examples
"""

# Imports
import os
import sys
from collections.abc import Callable

from ..decorators import LogLevels, handle_error


def update_pip_and_required_packages() -> int:
	""" Update pip and required packages.

	Returns:
		int: Return code of the os.system call.
	"""
	return os.system(f"{sys.executable} -m pip install --upgrade pip setuptools build twine pkginfo packaging")

def build_package() -> int:
	""" Build the package.

	Returns:
		int: Return code of the os.system call.
	"""
	return os.system(f"{sys.executable} -m build")

def upload_package(repository: str, filepath: str) -> int:
	""" Upload the package to PyPI.

	Args:
		repository  (str): Repository to upload to.
		filepath    (str): Path to the file to upload.

	Returns:
		int: Return code of the os.system call.
	"""
	return os.system(f"{sys.executable} -m twine upload --verbose -r {repository} {filepath}")

@handle_error(message="Error while doing the pypi full routine", error_log=LogLevels.ERROR_TRACEBACK)
def pypi_full_routine(
	repository: str,
	dist_directory: str,
	last_files: int = 1,
	endswith: str = ".tar.gz",

	update_all_function: Callable[[], int] = update_pip_and_required_packages,
	build_package_function: Callable[[], int] = build_package,
	upload_package_function: Callable[[str, str], int] = upload_package,
) -> None:
	""" Upload the most recent file(s) to PyPI after updating pip and required packages and building the package.

	Args:
		repository               (str):                        Repository to upload to.
		dist_directory           (str):                        Directory to upload from.
		last_files               (int):                        Number of most recent files to upload. Defaults to 1.
		endswith                 (str):                        End of the file name to upload. Defaults to ".tar.gz".
		update_all_function      (Callable[[], int]):          Function to update pip and required packages.
			Defaults to :func:`update_pip_and_required_packages`.
		build_package_function   (Callable[[], int]):          Function to build the package.
			Defaults to :func:`build_package`.
		upload_package_function  (Callable[[str, str], int]):  Function to upload the package.
			Defaults to :func:`upload_package`.

	Returns:
		int: Return code of the command.
	"""
	if update_all_function() != 0:
		raise Exception("Error while updating pip and required packages")

	if build_package_function() != 0:
		raise Exception("Error while building the package")

	# Get list of tar.gz files in dist directory sorted by modification time
	files: list[str] = sorted(
		[x for x in os.listdir(dist_directory) if x.endswith(endswith)],	# Get list of tar.gz files in dist directory
		key=lambda x: os.path.getmtime(f"{dist_directory}/{x}"),			# Sort by modification time
		reverse=True														# Sort in reverse order
	)

	# Upload the most recent file(s)
	for file in files[:last_files]:
		upload_package_function(repository, f"{dist_directory}/{file}")





""" Sandbox model implementation. (Where I try strange things)

Tested:

- ConvNeXtBase with input_shape=(1024, 1024, 3)
- Custom CNN architecture for implant classification (fixed / not fixed)

"""

# pyright: reportUnknownArgumentType=false
# pyright: reportUnknownVariableType=false
# pyright: reportMissingTypeStubs=false

# Imports
from __future__ import annotations

from keras.layers import (
	BatchNormalization,
	Conv2D,
	Input,
	MaxPooling2D,
	SpatialDropout2D,
)
from keras.models import Model

from ...print import warning
from ...decorators import simple_cache
from .base_keras import BaseKeras
from .model_interface import CLASS_ROUTINE_DOCSTRING, MODEL_DOCSTRING


class Sandbox(BaseKeras):
	def _get_base_model(self) -> Model:
		return self.custom_architecture()

	def custom_architecture(self) -> Model:
		""" Create a custom architecture for implant classification.

        This model uses a series of convolutional blocks with increasing depth,
        batch normalization, spatial and regular dropout for regularization.
        It's designed to detect features relevant to implant fixation status.

        Note: This is a custom architecture that does not use transfer learning.
        The transfer_learning attribute is ignored.

        Returns:
            Model: A Keras model without top layers for implant classification
        """
		if self.transfer_learning != "":
			warning(
				f"Transfer learning '{self.transfer_learning}' specified but not supported for custom architecture. "
				f"Using a model trained from scratch instead."
			)

		# Default input shape based on dataset loading defaults (224x224x3)
		input_shape: tuple[int, int, int] = (224, 224, 3)

		# Input layer
		inputs = Input(shape=input_shape)

		# Block 1: Initial feature extraction
		x = Conv2D(64, (3, 3), activation="relu", padding="same", name="block1_conv1")(inputs)
		x = BatchNormalization()(x)
		x = Conv2D(64, (3, 3), activation="relu", padding="same", name="block1_conv2")(x)
		x = BatchNormalization()(x)
		x = MaxPooling2D((2, 2), strides=(2, 2), name="block1_pool")(x)
		x = SpatialDropout2D(0.1)(x)

		# Block 2: Intermediate features
		x = Conv2D(128, (3, 3), activation="relu", padding="same", name="block2_conv1")(x)
		x = BatchNormalization()(x)
		x = Conv2D(128, (3, 3), activation="relu", padding="same", name="block2_conv2")(x)
		x = BatchNormalization()(x)
		x = MaxPooling2D((2, 2), strides=(2, 2), name="block2_pool")(x)
		x = SpatialDropout2D(0.1)(x)

		# Block 3: More complex features
		x = Conv2D(256, (3, 3), activation="relu", padding="same", name="block3_conv1")(x)
		x = BatchNormalization()(x)
		x = Conv2D(256, (3, 3), activation="relu", padding="same", name="block3_conv2")(x)
		x = BatchNormalization()(x)
		x = Conv2D(256, (3, 3), activation="relu", padding="same", name="block3_conv3")(x)
		x = BatchNormalization()(x)
		x = MaxPooling2D((2, 2), strides=(2, 2), name="block3_pool")(x)
		x = SpatialDropout2D(0.1)(x)

		# Block 4: Deep features
		x = Conv2D(512, (3, 3), activation="relu", padding="same", name="block4_conv1")(x)
		x = BatchNormalization()(x)
		x = Conv2D(512, (3, 3), activation="relu", padding="same", name="block4_conv2")(x)
		x = BatchNormalization()(x)
		x = Conv2D(512, (3, 3), activation="relu", padding="same", name="block4_conv3")(x)
		x = BatchNormalization()(x)
		x = MaxPooling2D((2, 2), strides=(2, 2), name="block4_pool")(x)
		x = SpatialDropout2D(0.1)(x)

		# Block 5: High-level abstract features
		x = Conv2D(512, (3, 3), activation="relu", padding="same", name="block5_conv1")(x)
		x = BatchNormalization()(x)
		x = Conv2D(512, (3, 3), activation="relu", padding="same", name="block5_conv2")(x)
		x = BatchNormalization()(x)
		x = Conv2D(512, (3, 3), activation="relu", padding="same", name="block5_conv3")(x)
		x = BatchNormalization()(x)

		# Create the model
		model = Model(inputs, x, name="implant_classifier")

		return model


# Docstrings
for model in [Sandbox]:
	model.__doc__ = MODEL_DOCSTRING.format(model=model.__name__)
	model.class_routine = simple_cache(model.class_routine)
	model.class_routine.__doc__ = CLASS_ROUTINE_DOCSTRING.format(model=model.__name__)





# pyright: reportUnusedImport=false
# ruff: noqa: F401

# Imports
from .common import Any, NDArray, check_image, cv2, np


# Functions
def zoom_image(image: NDArray[Any], zoom_factor: float, ignore_dtype: bool = False) -> NDArray[Any]:
	""" Zoom into an image.

	Args:
		image         (NDArray[Any]):  Image to zoom
		zoom_factor   (float):       Zoom factor (greater than 1 for zoom in, less than 1 for zoom out)
		ignore_dtype  (bool):        Ignore the dtype check
	Returns:
		NDArray[Any]: Zoomed image

	>>> ## Basic tests
	>>> image = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
	>>> zoomed = zoom_image(image.astype(np.uint8), 1.5)
	>>> zoomed.shape == image.shape
	True

	>>> img = np.eye(4, dtype=np.uint8) * 255
	>>> zoomed_in = zoom_image(img, 2.0)
	>>> zoomed_in.shape == img.shape  # Should preserve size
	True

	>>> zoomed_out = zoom_image(img, 0.5)
	>>> zoomed_out.shape == img.shape  # Should preserve size
	True

	>>> rgb = np.full((4,4,3), 128, dtype=np.uint8)
	>>> zoomed_rgb = zoom_image(rgb, 1.5)
	>>> zoomed_rgb.shape == (4,4,3)
	True

	>>> ## Test invalid inputs
	>>> zoom_image("not an image", 1.5)
	Traceback (most recent call last):
		...
	AssertionError: Image must be a numpy array

	>>> zoom_image(image.astype(np.uint8), "1.5")
	Traceback (most recent call last):
		...
	AssertionError: zoom_factor must be a number, got <class 'str'>

	>>> zoom_image(image.astype(np.uint8), -1)
	Traceback (most recent call last):
		...
	AssertionError: zoom_factor must be greater than 0, got -1
	"""
	# Check input data
	check_image(image, ignore_dtype=ignore_dtype)
	assert isinstance(zoom_factor, float | int), f"zoom_factor must be a number, got {type(zoom_factor)}"
	assert zoom_factor > 0, f"zoom_factor must be greater than 0, got {zoom_factor}"

	# Get image dimensions
	height, width = image.shape[:2]

	# Calculate new dimensions
	new_height, new_width = int(height * zoom_factor), int(width * zoom_factor)

	# Resize image
	zoomed_image: NDArray[Any] = cv2.resize(image, (new_width, new_height))

	# Crop or pad to original size
	if zoom_factor > 1:
		# Crop
		start_x: int = (new_width - width) // 2
		start_y: int = (new_height - height) // 2
		return zoomed_image[start_y:start_y + height, start_x:start_x + width] # pyright: ignore [reportUnknownVariableType]
	else:
		# Pad
		pad_x: int = (width - new_width) // 2
		pad_y: int = (height - new_height) // 2
		# Ensure value list matches number of channels (max 4 for OpenCV)
		value: list[int] = [0] * min(image.shape[-1], 4)
		return cv2.copyMakeBorder(zoomed_image, pad_y, pad_y, pad_x, pad_x, cv2.BORDER_CONSTANT, value=value)




""" Keras utilities for generating Grad-CAM heatmaps and saliency maps for model interpretability. """

# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false
# pyright: reportUnknownArgumentType=false
# pyright: reportMissingTypeStubs=false

# Imports
import os
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
from keras.models import Model
from matplotlib.colors import Colormap
from numpy.typing import NDArray
from PIL import Image

from ....decorators import handle_error
from ....print import error, warning
from ...config.get import DataScienceConfig


@handle_error(error_log=DataScienceConfig.ERROR_LOG)
def make_gradcam_heatmap(
	model: Model,
	img: NDArray[Any],
	class_idx: int = 0,
	last_conv_layer_name: str = "",
	one_per_channel: bool = False
) -> list[NDArray[Any]]:
	""" Generate a Grad-CAM heatmap for a given image and model.

	Args:
		model                 (Model):       The pre-trained TensorFlow model
		img                   (NDArray[Any]):  The preprocessed image array (ndim=3 or 4 with shape=(1, ?, ?, ?))
		class_idx             (int):         The class index to use for the Grad-CAM heatmap
		last_conv_layer_name  (str):         Name of the last convolutional layer in the model
			(optional, will try to find it automatically)
		one_per_channel       (bool):        If True, return one heatmap per channel
	Returns:
		list[NDArray[Any]]: The Grad-CAM heatmap(s)

	Examples:
		.. code-block:: python

			> model: Model = ...
			> img: NDArray[Any] = np.array(Image.open("path/to/image.jpg").convert("RGB"))
			> last_conv_layer: str = Utils.find_last_conv_layer(model) or "conv5_block3_out"
			> heatmap: NDArray[Any] = Utils.make_gradcam_heatmap(model, img, last_conv_layer)[0]
			> Image.fromarray(heatmap).save("heatmap.png")
	"""
	# Assertions
	assert isinstance(model, Model), "Model must be a valid Keras model"
	assert isinstance(img, np.ndarray), "Image array must be a valid numpy array"

	# If img is not a batch of 1, convert it to a batch of 1
	if img.ndim == 3:
		img = np.expand_dims(img, axis=0)
	assert img.ndim == 4 and img.shape[0] == 1, "Image array must be a batch of 1 (shape=(1, ?, ?, ?))"

	# If last_conv_layer_name is not provided, find it automatically
	if not last_conv_layer_name:
		last_conv_layer_name = find_last_conv_layer(model)
		if not last_conv_layer_name:
			error("Last convolutional layer not found. Please provide the name of the last convolutional layer.")

	# Get the last convolutional layer
	last_layer: Model | list[Model] = model.get_layer(last_conv_layer_name)
	if isinstance(last_layer, list):
		last_layer = last_layer[0]

	# Create a model that outputs both the last conv layer's activations and predictions
	grad_model: Model = Model(
		[model.inputs],
		[last_layer.output, model.output]
	)

	# Record operations for automatic differentiation using GradientTape.
	with tf.GradientTape() as tape:
		# Forward pass: get the activations of the last conv layer and the predictions.
		conv_outputs: tf.Tensor
		predictions: tf.Tensor
		conv_outputs, predictions = grad_model(img)
		# print("conv_outputs shape:", conv_outputs.shape)
		# print("predictions shape:", predictions.shape)

		# Compute the loss with respect to the positive class
		loss: tf.Tensor = predictions[:, class_idx]

		# Compute the gradient of the loss with respect to the activations ..
		# .. of the last conv layer
		grads: tf.Tensor = tf.convert_to_tensor(tape.gradient(loss, conv_outputs))
		# print("grads shape:", grads.shape)

	# Initialize heatmaps list
	heatmaps: list[NDArray[Any]] = []

	if one_per_channel:
		# Return one heatmap per channel
		# Get the number of filters in the last conv layer
		num_filters: int = int(tf.shape(conv_outputs)[-1])

		for i in range(num_filters):
			# Compute the mean intensity of the gradients for this filter
			pooled_grad: tf.Tensor = tf.reduce_mean(grads[:, :, :, i])

			# Get the activation map for this filter
			activation_map: tf.Tensor = conv_outputs[:, :, :, i]

			# Weight the activation map by the gradient
			weighted_map: tf.Tensor = activation_map * pooled_grad

			# Ensure that the heatmap has non-negative values and normalize it
			heatmap: tf.Tensor = tf.maximum(weighted_map, 0) / tf.math.reduce_max(weighted_map)

			# Remove possible dims of size 1 + convert the heatmap to a numpy array
			heatmaps.append(tf.squeeze(heatmap).numpy())
	else:
		# Compute the mean intensity of the gradients for each filter in the last conv layer.
		pooled_grads: tf.Tensor = tf.reduce_mean(grads, axis=(0, 1, 2))

		# Multiply each activation map in the last conv layer by the corresponding
		# gradient average, which emphasizes the parts of the image that are important.
		heatmap: tf.Tensor = conv_outputs @ tf.expand_dims(pooled_grads, axis=-1)
		# print("heatmap shape (before squeeze):", heatmap.shape)

		# Ensure that the heatmap has non-negative values and normalize it.
		heatmap = tf.maximum(heatmap, 0) / tf.math.reduce_max(heatmap)

		# Remove possible dims of size 1 + convert the heatmap to a numpy array.
		heatmaps.append(tf.squeeze(heatmap).numpy())

	return heatmaps

@handle_error(error_log=DataScienceConfig.ERROR_LOG)
def make_saliency_map(
	model: Model,
	img: NDArray[Any],
	class_idx: int = 0,
	one_per_channel: bool = False
) -> list[NDArray[Any]]:
	""" Generate a saliency map for a given image and model.

	A saliency map shows which pixels in the input image have the greatest influence on the
	model's prediction.

	Args:
		model            (Model):       The pre-trained TensorFlow model
		img              (NDArray[Any]):  The preprocessed image array (batch of 1)
		class_idx        (int):         The class index to use for the saliency map
		one_per_channel  (bool):        If True, return one saliency map per channel
	Returns:
		list[NDArray[Any]]: The saliency map(s) normalized to range [0,1]

	Examples:
		.. code-block:: python

			> model: Model = ...
			> img: NDArray[Any] = np.array(Image.open("path/to/image.jpg").convert("RGB"))
			> saliency: NDArray[Any] = Utils.make_saliency_map(model, img)[0]
	"""
	# Assertions
	assert isinstance(model, Model), "Model must be a valid Keras model"
	assert isinstance(img, np.ndarray), "Image array must be a valid numpy array"

	# If img is not a batch of 1, convert it to a batch of 1
	if img.ndim == 3:
		img = np.expand_dims(img, axis=0)
	assert img.ndim == 4 and img.shape[0] == 1, "Image array must be a batch of 1 (shape=(1, ?, ?, ?))"

	# Convert the numpy array to a tensor
	img_tensor = tf.convert_to_tensor(img, dtype=tf.float32)

	# Record operations for automatic differentiation
	with tf.GradientTape() as tape:
		tape.watch(img_tensor)
		predictions = model(img_tensor)
		# Use class index for positive class prediction
		loss = predictions[:, class_idx]

	# Compute gradients of loss with respect to input image
	try:
		grads = tape.gradient(loss, img_tensor)
		# Cast to tensor to satisfy type checker
		grads = tf.convert_to_tensor(grads)
	except Exception as e:
		warning(f"Error computing gradients: {e}")
		return []

	# Initialize saliency maps list
	saliency_maps: list[NDArray[Any]] = []

	if one_per_channel:
		# Return one saliency map per channel
		# Get the last dimension size as an integer
		num_channels: int = int(tf.shape(grads)[-1])
		for i in range(num_channels):
			# Extract gradients for this channel
			channel_grads: tf.Tensor = tf.abs(grads[:, :, :, i])

			# Apply smoothing to make the map more visually coherent
			channel_grads = tf.nn.avg_pool2d(
				tf.expand_dims(channel_grads, -1),
				ksize=3, strides=1, padding='SAME'
			)
			channel_grads = tf.squeeze(channel_grads)

			# Normalize the saliency map for this channel
			channel_max = tf.math.reduce_max(channel_grads)
			if channel_max > 0:
				channel_saliency: tf.Tensor = channel_grads / channel_max
			else:
				channel_saliency: tf.Tensor = channel_grads

			saliency_maps.append(channel_saliency.numpy().squeeze())
	else:
		# Take absolute value of gradients
		abs_grads = tf.abs(grads)

		# Sum across color channels for RGB images
		saliency = tf.reduce_sum(abs_grads, axis=-1)

		# Apply smoothing to make the map more visually coherent
		saliency = tf.nn.avg_pool2d(
			tf.expand_dims(saliency, -1),
			ksize=3, strides=1, padding='SAME'
		)
		saliency = tf.squeeze(saliency)

		# Normalize saliency map
		saliency_max = tf.math.reduce_max(saliency)
		if saliency_max > 0:
			saliency = saliency / saliency_max

		saliency_maps.append(saliency.numpy().squeeze())

	return saliency_maps

@handle_error(error_log=DataScienceConfig.ERROR_LOG)
def find_last_conv_layer(model: Model) -> str:
	""" Find the name of the last convolutional layer in a model.

	Args:
		model (Model): The TensorFlow model to analyze
	Returns:
		str: Name of the last convolutional layer if found, otherwise an empty string

	Examples:
		.. code-block:: python

			> model: Model = ...
			> last_conv_layer: str = Utils.find_last_conv_layer(model)
			> print(last_conv_layer)
			'conv5_block3_out'
	"""
	assert isinstance(model, Model), "Model must be a valid Keras model"

	# Find the last convolutional layer by iterating through the layers in reverse
	last_conv_layer_name: str = ""
	for layer in reversed(model.layers):
		if isinstance(layer, tf.keras.layers.Conv2D):
			last_conv_layer_name = layer.name
			break

	# Return the name of the last convolutional layer
	return last_conv_layer_name

@handle_error(error_log=DataScienceConfig.ERROR_LOG)
def create_visualization_overlay(
	original_img: NDArray[Any] | Image.Image,
	heatmap: NDArray[Any],
	alpha: float = 0.4,
	colormap: str = "jet"
) -> NDArray[Any]:
	""" Create an overlay of the original image with a heatmap visualization.

	Args:
		original_img  (NDArray[Any] | Image.Image):  The original image array or PIL Image
		heatmap       (NDArray[Any]):                The heatmap to overlay (normalized to 0-1)
		alpha         (float):                     Transparency level of overlay (0-1)
		colormap      (str):                       Matplotlib colormap to use for heatmap
	Returns:
		NDArray[Any]: The overlaid image

	Examples:
		.. code-block:: python

			> original: NDArray[Any] | Image.Image = ...
			> heatmap: NDArray[Any] = Utils.make_gradcam_heatmap(model, img)[0]
			> overlay: NDArray[Any] = Utils.create_visualization_overlay(original, heatmap)
			> Image.fromarray(overlay).save("overlay.png")
	"""
	# Ensure heatmap is normalized to 0-1
	if heatmap.max() > 0:
		heatmap = heatmap / heatmap.max()

	## Apply colormap to heatmap
	# Get the colormap
	cmap: Colormap = plt.cm.get_cmap(colormap)

	# Ensure heatmap is at least 2D to prevent tuple return from cmap
	if np.isscalar(heatmap) or heatmap.ndim == 0:
		heatmap = np.array([[heatmap]])

	# Apply colormap and ensure it's a numpy array
	heatmap_rgb: NDArray[Any] = np.array(cmap(heatmap))

	# Remove alpha channel if present
	if heatmap_rgb.ndim >= 3:
		heatmap_rgb = heatmap_rgb[:, :, :3]

	heatmap_rgb = (255 * heatmap_rgb).astype(np.uint8)

	# Convert original image to PIL Image if it's a numpy array
	if isinstance(original_img, np.ndarray):
		# Scale to 0-255 range if image is normalized (0-1)
		if original_img.max() <= 1:
			img: NDArray[Any] = (original_img * 255).astype(np.uint8)
		else:
			img: NDArray[Any] = original_img.astype(np.uint8)
		original_pil: Image.Image = Image.fromarray(img)
	else:
		original_pil: Image.Image = original_img

	# Convert heatmap to PIL Image
	heatmap_pil: Image.Image = Image.fromarray(heatmap_rgb)

	# Resize heatmap to match original image size if needed
	if heatmap_pil.size != original_pil.size:
		heatmap_pil = heatmap_pil.resize(original_pil.size, Image.Resampling.LANCZOS)

	# Ensure both images have the same mode before blending
	if original_pil.mode != heatmap_pil.mode:
		if original_pil.mode == "RGB" and heatmap_pil.mode == "RGBA":
			heatmap_pil = heatmap_pil.convert("RGB")
		else:
			original_pil = original_pil.convert(heatmap_pil.mode)

	# Blend images
	overlaid_img: Image.Image = Image.blend(original_pil, heatmap_pil, alpha=alpha)

	return np.array(overlaid_img)

@handle_error(error_log=DataScienceConfig.ERROR_LOG)
def all_visualizations_for_image(
	model: Model,
	folder_path: str,
	img: NDArray[Any],
	base_name: str,
	class_idx: int,
	class_name: str,
	files: tuple[str, ...],
	data_type: str
) -> None:
	""" Process a single image to generate visualizations and determine prediction correctness.

	Args:
		model          (Model):           The pre-trained TensorFlow model
		folder_path    (str):             The path to the folder where the visualizations will be saved
		img            (NDArray[Any]):    The preprocessed image array (batch of 1)
		base_name      (str):             The base name of the image
		class_idx      (int):             The **true** class index for the image
		class_name     (str):             The **true** class name for organizing folders
		files          (tuple[str, ...]): List of original file paths for the subject
		data_type      (str):             Type of data ("test" or "train")
	"""
	# Set directories based on data type
	saliency_dir: str = f"{folder_path}/{data_type}/saliency_maps"
	grad_cam_dir: str = f"{folder_path}/{data_type}/grad_cams"

	# Convert label to class name and load all original images
	original_imgs: list[Image.Image] = [Image.open(f).convert("RGB") for f in files]

	# Find the maximum dimensions across all images
	max_shape: tuple[int, int] = max(i.size for i in original_imgs)

	# Resize all images to match the largest dimensions while preserving aspect ratio
	resized_imgs: list[NDArray[Any]] = []
	for original_img in original_imgs:
		if original_img.size != max_shape:
			original_img = original_img.resize(max_shape, resample=Image.Resampling.LANCZOS)
		resized_imgs.append(np.array(original_img))

	# Take mean of resized images
	subject_image: NDArray[Any] = np.mean(np.stack(resized_imgs), axis=0).astype(np.uint8)

	# Perform prediction to determine correctness
	# Ensure img is in batch format (add dimension if needed)
	img_batch: NDArray[Any] = np.expand_dims(img, axis=0) if img.ndim == 3 else img
	predicted_class_idx: int = int(np.argmax(model.predict(img_batch, verbose=0)[0])) # pyright: ignore [reportArgumentType]
	prediction_correct: bool = (predicted_class_idx == class_idx)
	status_suffix: str = "_correct" if prediction_correct else "_missed"

	# Create and save Grad-CAM visualization
	heatmap: NDArray[Any] = make_gradcam_heatmap(model, img, class_idx=class_idx, one_per_channel=False)[0]
	overlay: NDArray[Any] = create_visualization_overlay(subject_image, heatmap)

	# Save the overlay visualization with status suffix
	grad_cam_path = f"{grad_cam_dir}/{class_name}/{base_name}{status_suffix}_gradcam.png"
	os.makedirs(os.path.dirname(grad_cam_path), exist_ok=True)
	Image.fromarray(overlay).save(grad_cam_path)

	# Create and save saliency maps
	saliency_map: NDArray[Any] = make_saliency_map(model, img, class_idx=class_idx, one_per_channel=False)[0]
	overlay: NDArray[Any] = create_visualization_overlay(subject_image, saliency_map)

	# Save the overlay visualization with status suffix
	saliency_path = f"{saliency_dir}/{class_name}/{base_name}{status_suffix}_saliency.png"
	os.makedirs(os.path.dirname(saliency_path), exist_ok=True)
	Image.fromarray(overlay).save(saliency_path)





# pyright: reportUnusedImport=false
# ruff: noqa: F401

# Imports
from .common import Any, NDArray, check_image, cv2, np


# Functions
def auto_contrast_image(image: NDArray[Any], ignore_dtype: bool = False) -> NDArray[Any]:
	""" Adjust the contrast of an image.

	Args:
		image         (NDArray[Any]):  Image to adjust contrast
		ignore_dtype  (bool):        Ignore the dtype check
	Returns:
		NDArray[Any]: Image with adjusted contrast

	>>> ## Basic tests
	>>> image = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=np.uint8)
	>>> adjusted = auto_contrast_image(image)
	>>> adjusted.tolist()
	[[0, 36, 73], [109, 146, 182], [219, 255, 255]]
	>>> adjusted.shape == image.shape
	True
	>>> adjusted.dtype == image.dtype
	True

	>>> ## Test invalid inputs
	>>> auto_contrast_image("not an image")
	Traceback (most recent call last):
		...
	AssertionError: Image must be a numpy array
	"""
	# Check input data
	check_image(image, ignore_dtype=ignore_dtype)

	# Perform histogram clipping
	clip_hist_percent: float = 1.0

	# Calculate the histogram of the image
	hist: NDArray[Any] = cv2.calcHist([image], [0], None, [256], [0, 256])

	# Create an accumulator list to store the cumulative histogram
	accumulator: list[float] = []
	accumulator.append(hist[0])
	for i in range(1, 256):
		accumulator.append(accumulator[i - 1] + hist[i])

	# Find the maximum value in the accumulator
	max_value: float = accumulator[-1]

	# Calculate the clipping threshold
	clip_hist_percent = clip_hist_percent * (max_value / 100.0)
	clip_hist_percent = clip_hist_percent / 2.0

	# Find the minimum and maximum gray levels after clipping
	min_gray: int = 0
	while accumulator[min_gray] < clip_hist_percent:
		min_gray = min_gray + 1
	max_gray: int = 256 - 1
	while (max_gray >= 0 and accumulator[max_gray] >= (max_value - clip_hist_percent)):
		max_gray = max_gray - 1

	# Calculate the input range after clipping
	input_range: int = max_gray - min_gray

	# If the input range is 0, return the original image
	if input_range == 0:
		return image

	# Calculate the scaling factors for contrast adjustment
	alpha: float = (256 - 1) / input_range
	beta: float = -min_gray * alpha

	# Apply the contrast adjustment
	adjusted: NDArray[Any] = cv2.convertScaleAbs(image, alpha=alpha, beta=beta)
	return adjusted




""" Configuration file for the project. """

# Imports
import os
from typing import Literal

from stouputils.decorators import LogLevels
from stouputils.io import get_root_path

# Environment variables
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "9"         # Suppress TensorFlow logging
os.environ["GRPC_VERBOSITY"] = "ERROR"           # Suppress gRPC logging


# Configuration class
class DataScienceConfig:
	""" Configuration class for the project. """

	# Common
	SEED: int = 42
	""" Seed for the random number generator. """

	ERROR_LOG: LogLevels = LogLevels.WARNING_TRACEBACK
	""" Log level for errors for all functions. """

	AUGMENTED_FILE_SUFFIX: str = "_aug_"
	""" Suffix for augmented files, ex: 'image_008_aug_1.png'. """

	AUGMENTED_DIRECTORY_PREFIX: str = "aug_"
	""" Prefix for augmented directories, ex: 'data/hip_implant' -> 'data/aug_hip_implant'. """

	PREPROCESSED_DIRECTORY_SUFFIX: str = "_preprocessed"
	""" Suffix for preprocessed directories, ex: 'data/hip_implant' -> 'data/hip_implant_preprocessed'. """


	# Directories
	ROOT: str = get_root_path(__file__, go_up=3)
	""" Root directory of the project. """

	MLFLOW_FOLDER: str = f"{ROOT}/mlruns"
	""" Folder containing the mlflow data. """
	MLFLOW_URI: str = f"file://{MLFLOW_FOLDER}"
	""" URI to the mlflow data. """

	DATA_FOLDER: str = f"{ROOT}/data"
	""" Folder containing all the data (e.g. subfolders containing images). """

	TEMP_FOLDER: str = f"{ROOT}/temp"
	""" Folder containing temporary files (e.g. models checkpoints, plots, etc.). """

	LOGS_FOLDER: str = f"{ROOT}/logs"
	""" Folder containing the logs. """

	TENSORBOARD_FOLDER: str = f"{ROOT}/tensorboard"
	""" Folder containing the tensorboard logs. """


	# Behaviours
	TEST_SIZE: float = 0.2
	""" Size of the test set by default (0.2 means 80% training, 20% test). """

	VALIDATION_SIZE: float = 0.2
	""" Size of the validation set by default (0.2 means 80% training, 20% validation). """

	# Machine learning
	SAVE_MODEL: bool = False
	""" If the model should be saved in the mlflow folder using mlflow.*.save_model. """

	DO_SALIENCY_AND_GRADCAM: bool = True
	""" If the saliency and gradcam should be done during the run. """

	DO_LEARNING_RATE_FINDER: Literal[0, 1, 2] = 1
	""" If the learning rate finder should be done during the run.
	0: no, 1: only plot, 2: plot and use value for the remaining run
	"""

	DO_UNFREEZE_FINDER: Literal[0, 1, 2] = 0
	""" If the unfreeze finder should be done during the run.
	0: no, 1: only plot, 2: plot and use value for the remaining run
	"""

	DO_FIT_IN_SUBPROCESS: bool = True
	""" If the model should be fitted in a subprocess.
	Is memory efficient, and more stable. Turn it off only if you are having issues.

	Note: This allow a program to make lots of runs without getting killed by the OS for using too much resources.
	(e.g. LeaveOneOut Cross Validation, Grid Search, etc.)
	"""

	MIXED_PRECISION_POLICY: Literal["mixed_float16", "mixed_bfloat16", "float32"] = "mixed_float16"
	""" Mixed precision policy to use. Turn back to "float32" if you are having issues with a specific model or metrics.
	See: https://www.tensorflow.org/guide/mixed_precision
	"""

	TENSORFLOW_DEVICE: str = "/gpu:1"
	""" TensorFlow device to use. """



# Fix MLFLOW_URI for Windows by adding a missing slash
if os.name == "nt":
	DataScienceConfig.MLFLOW_URI = DataScienceConfig.MLFLOW_URI.replace("file:", "file:/")

